#!/usr/bin/env python3

"""metaTOR - a set of scripts that streamlines the processing and binning of
metagenomic 3C datasets.
"""


import subprocess
import sys
import pkg_resources
import pathlib
import requests
import tarfile

BASE_PRODIGAL = (
    "https://github.com/hyattpd/Prodigal/releases/download/v2.6.3/prodigal"
)
HMM_URL = "http://dl.pasteur.fr/fop/LItxiFe9/hmm_databases.tgz"


def download_and_install_dependencies():
    """Setup URLS and download dependencies.
    """

    dependencies = {"hmm_databases": HMM_URL}

    if sys.platform.startswith("linux") or "bsd" in sys.platform:

        dependencies["prodigal"] = "{}.linux".format(BASE_PRODIGAL)
        dependencies["louvain"] = (
            "https://lip6.github.io/Louvain-BinaryBuild/"
            "louvain_linux.tar.gz"
        )
    elif sys.platform == "darwin":

        dependencies["prodigal"] = "{}.osx.10.9.5".format(BASE_PRODIGAL)
        dependencies["louvain"] = (
            "https://github.com/lip6/Louvain-BinaryBuilds/raw/osx/"
            "louvain_osx.tar.gz"
        )
    elif sys.platform.startswith("win") or sys.platform == "cygwin":

        dependencies["prodigal"] = "{}.windows.exe"
        dependencies["louvain"] = (
            "https://ci.appveyor.com/api/projects/yanntm/"
            "Louvain-BinaryBuild/artifacts/website/"
            "louvain_windows.tar.gz"
        )

    else:
        raise NotImplementedError(
            "Your platform is not supported: {}".format(sys.platform)
        )

    cache_dir = pathlib.Path.cwd() / pathlib.Path("cache")

    try:
        print("Downloading dependencies...")
        cache_dir.mkdir()
        for dependency_name, url in dependencies.items():
            print("Downloading {} at {}".format(dependency_name, url))
            request = requests.get(url)
            basename = url.split("/")[-1]

            with open(cache_dir / basename, "wb") as handle:
                print(dependency_name, basename, cache_dir / basename)
                handle.write(request.content)
    except FileExistsError:
        print("Using cached dependencies...")

    share_dir = pathlib.Path.cwd()
    tools_dir = share_dir / "tools"
    louvain_dir = tools_dir / "louvain"

    louvain_dir.mkdir(parents=True, exist_ok=True)

    louvain_basename = dependencies["louvain"].split("/")[-1]
    louvain_path = louvain_dir / louvain_basename
    (cache_dir / louvain_basename).replace(louvain_path)

    with tarfile.open(louvain_path, "r:gz") as tar:
        tar.extractall()

    hmm_basename = dependencies["hmm_databases"].split("/")[-1]
    hmm_path = share_dir / hmm_basename
    (cache_dir / hmm_basename).replace(hmm_path)

    prodigal_basename = dependencies["prodigal"].split("/")[-1]
    prodigal_path = tools_dir / "prodigal"
    (cache_dir / prodigal_basename).replace(prodigal_path)


def main():
    """This module just acts as an entry point to the bulk of the pipeline.
    All argument parsing is delegated to metator.sh
    """

    metator_args = sys.argv[1:]
    entry_point = pkg_resources.resource_filename("metator", "bin/metator.sh")

    metator_process = subprocess.Popen((entry_point, *metator_args))
    metator_process.wait()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Process metagenomic bins and their networks

Extract bin data and figures from partitions, namely bin subnetworks
(notably for recursion purposes) and bin fasta files. Also draws the subnetwork
in adjacency matrix form.

"""

from __future__ import absolute_import

import numpy as np
from scipy import sparse
from scipy import stats
from Bio import SeqIO
import os
import argparse
import itertools
import operator
from metator.scripts import hicstuff as hcs

from metator.scripts.log import logger

from metator.scripts.figures import spaceless_pdf_plot_maker

from metator.scripts.network import DEFAULT_CHUNK_SIZE
from metator.scripts.figures import DEFAULT_MAX_SIZE_MATRIX
from metator.scripts.figures import DEFAULT_SATURATION_THRESHOLD

DEFAULT_MAX_CORES = 100


def extract_subnetworks(
    partition_file,
    network_file,
    output_dir,
    max_cores=DEFAULT_MAX_CORES,
    max_size_matrix=DEFAULT_MAX_SIZE_MATRIX,
    saturation_threshold=DEFAULT_SATURATION_THRESHOLD,
):
    """Extract bin subnetworks from the main network

    Identify bins, extract subnets, draws the adjacency matrices,
    saves it all in a specified output directory.

    Parameters
    ----------
    partition_file : file, str or pathlib.Path
        The file containing, for each chunk, the communities it was
        assigned to at each iteration.
    network_file : file, str or pathlib.Path
        The file containing the network in sparse (edge list) format
    output_dir : str or pathlib.Path
        The output directory to write the subnetworks into.
    max_cores : int, optional
        The maximum number of bins to extract. Default is 100.
    max_size_matrix : int, optional
        When rendering contact maps for each bin, the maximum size for the
        matrix. Default is 2000.
    saturation_threshold : float, optional
        When rendering contact maps for each bin, the percentile value over
        which the color map should be saturated. Default is 80.

    Returns
    -------
    None
    """

    logger.info("Loading partition...")
    data_chunks = np.loadtxt(partition_file, usecols=(1,), dtype=np.int32)

    logger.info("Loading network...")
    network = np.loadtxt(network_file, dtype=np.int32)
    cores = data_chunks

    core_network = np.copy(network)

    core_network[:, 0] = cores[network[:, 0]]
    core_network[:, 1] = cores[network[:, 1]]

    n = np.amax(cores) + 1

    def extract(network_to_keep, filename):

        subnetwork = np.copy(network[network_to_keep])
        subnetwork[:, 0] -= 1
        subnetwork[:, 1] -= 1

        np.savetxt(filename, subnetwork, fmt="%i")

        return subnetwork

    def draw(subnetwork, filename):

        try:
            # Numpy array format
            row = subnetwork[:, 0]
            col = subnetwork[:, 1]
            data = subnetwork[:, 2]
        except TypeError:
            # Scipy sparse format
            row = subnetwork.row
            col = subnetwork.col
            data = subnetwork.data

        row_indices = stats.rankdata(
            np.concatenate((row, col)), method="dense"
        )
        col_indices = stats.rankdata(
            np.concatenate((col, row)), method="dense"
        )
        data = np.concatenate((data, data))

        # print("Row length: {}, col length: {}, data length: {}"
        #       "".format(len(row_indices), len(col_indices), len(data)))

        unique_row = np.unique(row)
        unique_col = np.unique(col)

        # print("Network shape: {},{}".format(len(unique_row),
        #                                     len(unique_col)))

        size = len(np.unique(np.concatenate((unique_row, unique_col)))) + 1
        # print("Size of matrix to draw: {}".format(size))

        try:
            sparse_subnet = sparse.coo_matrix(
                (data, (row_indices, col_indices)), shape=(size, size)
            )
            binning_factor = (size // max_size_matrix) + 1
            binned_subnet = hcs.bin_sparse(
                sparse_subnet, subsampling_factor=binning_factor
            )
            dense_subnet = binned_subnet.todense()

            diagonal = np.diag(np.diag(dense_subnet))
            normed_subnet = hcs.normalize_dense(dense_subnet - diagonal)

            vmax = np.percentile(normed_subnet, saturation_threshold)

            spaceless_pdf_plot_maker(normed_subnet, filename, vmax=vmax)

        except MemoryError:
            logger.warning(
                "Warning, couldn't save matrix due to memory issues"
            )

    def extract_and_draw(network_to_keep, filename_text, filename_image):

        subnetwork = extract(network_to_keep, filename=filename_text)
        draw(subnetwork, filename=filename_image)

    #   Extract and draw subnetworks for chosen cores and draw 2D arrays
    global_network_indices_list = []

    for i in range(1, n):

        if i > max_cores:
            break

        # print("Bin {}:".format(i))
        network_to_keep_1 = core_network[:, 0] == i
        network_to_keep_2 = core_network[:, 1] == i

        network_to_keep = network_to_keep_1 * network_to_keep_2

        nonzero_indices, = np.nonzero(network_to_keep)
        global_network_indices_list += nonzero_indices.tolist()

        subnetwork_file = os.path.join(
            output_dir, "subnetwork_core_{}.dat".format(i)
        )

        image_name = os.path.join(output_dir, "core_{}.eps".format(i))

        extract_and_draw(
            network_to_keep=network_to_keep,
            filename_text=subnetwork_file,
            filename_image=image_name,
        )


def extract_fasta(
    partition_file,
    fasta_file,
    output_dir,
    chunk_size=DEFAULT_CHUNK_SIZE,
    max_cores=DEFAULT_MAX_CORES,
):

    """Extract sequences from bins

    Identify bins, extract chunks belonging to each bins and gather them
    in a single FASTA file.

    Parameters
    ----------
    partition_file : file, str or pathlib.Path
        The file containing, for each chunk, the communities it was
        assigned to at each iteration.
    fasta_file : file, str or pathlib.Path
        The initial assembly from which chunks were initialized.
    output_dir : str or pathlib.Path
        The output directory to write the FASTA chunks into.
    chunk_size : int, optional
        The size of the chunks (in bp) used in the pipeline. Default is 1000.
    max_cores : int, optional
        How many bins to extract FASTA sequences from. Default is 100.

    Returns
    -------
    None
    """

    genome = {
        record.id: record.seq for record in SeqIO.parse(fasta_file, "fasta")
    }

    data_chunks = list(
        zip(*np.genfromtxt(partition_file, usecols=(0, 1), dtype=None))
    )

    chunk_names = np.array(data_chunks[0], dtype=object)
    cores = np.array(data_chunks[1])
    for core in set(cores):
        if core > max_cores:
            continue
        chunks_to_keep = chunk_names[cores == core]
        core_name = "core_{}.fa".format(core)

        core_file = os.path.join(output_dir, core_name)

        with open(core_file, "w") as core_handle:
            for name in chunks_to_keep:
                fields = name.split("_")
                header_name = "_".join(fields[:-1])
                chunk = int(fields[-1])

                pos_start = chunk * chunk_size
                pos_end = min(
                    (chunk + 1) * chunk_size, len(genome[header_name])
                )

                sequence = str(genome[header_name][pos_start:pos_end])

                core_handle.write(">{}\n".format(name))
                core_handle.write("{}\n".format(sequence))


def merge_fasta(fasta_file, output_dir):

    """Merge chunks into complete FASTA bins

    Merge bin chunks by appending consecutive chunks to one another.

    Parameters
    ----------
    fasta_file : file, str or pathlib.Path
        The FASTA file containing the chunks to merge.
    output_dir : str or pathlib.Path
        The output directory to write the merged FASTA bin into.

    Returns
    -------
    None
    """

    #   First, define some functions for ordering chunks and detecting
    #   consecutive chunk sequences

    def chunk_lexicographic_order(chunk):
        """A quick callback to sort chunk ids lexicographically
        (first on original names alphabetically, then on relative
        position on the original contig)
        """
        chunk_fields = chunk.split("_")
        chunk_name = chunk_fields[:-1]
        chunk_id = chunk_fields[-1]
        return (chunk_name, int(chunk_id))

    def are_consecutive(chunk1, chunk2):
        if None in {chunk1, chunk2}:
            return False
        else:
            ord1 = chunk_lexicographic_order(chunk1)
            ord2 = chunk_lexicographic_order(chunk2)
            return (ord1[0] == ord2[0]) and (ord1[1] == ord2[1] + 1)

    def consecutiveness(key_chunk_pair):
        """A callback for the groupby magic below
        """
        key, chunk = key_chunk_pair
        chunk_name, chunk_id = chunk_lexicographic_order(chunk)
        return (chunk_name, chunk_id - key)

    #   Read chunks and sort them
    genome = {
        record.id: record.seq for record in SeqIO.parse(fasta_file, "fasta")
    }

    sorted_ids = sorted(genome, key=chunk_lexicographic_order)

    #   Identify consecutive ranges and merge them
    new_genome = dict()
    for _, g in itertools.groupby(enumerate(sorted_ids), consecutiveness):
        chunk_range = map(operator.itemgetter(1), g)
        first_chunk = next(chunk_range)
        my_sequence = genome[first_chunk]
        my_chunk = None
        while "Reading chunk range":
            try:
                my_chunk = next(chunk_range)
                my_sequence += genome[my_chunk]
            except StopIteration:
                break

        try:
            last_chunk_id = my_chunk.split("_")[-1]
        except AttributeError:
            last_chunk_id = ""

        if last_chunk_id:
            new_chunk_id = "{}_{}".format(first_chunk, last_chunk_id)
        else:
            new_chunk_id = first_chunk
        new_genome[new_chunk_id] = my_sequence

    #   Write the result
    base_name = ".".join(os.path.basename(fasta_file).split(".")[:-1])
    output_name = "{}_merged.fa".format(base_name)
    merged_core_file = os.path.join(output_dir, output_name)
    with open(merged_core_file, "w") as output_handle:
        for my_id in sorted(new_genome, key=chunk_lexicographic_order):
            output_handle.write(">{}\n".format(my_id))
            output_handle.write("{}\n".format(new_genome[my_id]))


def main():

    parser = argparse.ArgumentParser(
        description="Extract bin data," " subnets and figures"
    )

    parser.add_argument(
        "-f", "--fasta", help="Reference FASTA file for extracting bins"
    )

    parser.add_argument(
        "-i", "--input", help="Input partition file for bin extraction"
    )

    parser.add_argument(
        "-o",
        "--output",
        required=True,
        help="Output directory for bin data and figures",
    )

    parser.add_argument(
        "-n", "--network", help="Reference network file for extracting subnets"
    )

    parser.add_argument(
        "-c",
        "--chunk-size",
        type=int,
        help="Chunk size for bin reconstruction",
        default=DEFAULT_CHUNK_SIZE,
    )

    parser.add_argument(
        "-m", "--merge", help="Merge consecutive chunks in a FASTA file"
    )

    parser.add_argument(
        "-N",
        "--n-bins",
        help="Number of bins to extract in size order",
        type=int,
        default=DEFAULT_MAX_CORES,
    )

    args = parser.parse_args()

    fasta_file = args.fasta
    partition_file = args.input
    output_dir = args.output
    network_file = args.network
    chunk_size = args.chunk_size
    fasta_to_merge = args.merge
    max_cores = args.n_bins

    if network_file is not None:
        extract_subnetworks(
            partition_file=partition_file,
            network_file=network_file,
            max_cores=max_cores,
            output_dir=output_dir,
        )

    elif fasta_file is not None:
        extract_fasta(
            partition_file=partition_file,
            fasta_file=fasta_file,
            output_dir=output_dir,
            max_cores=max_cores,
            chunk_size=chunk_size,
        )

    elif fasta_to_merge is not None:
        merge_fasta(fasta_file=fasta_to_merge, output_dir=output_dir)


if __name__ == "__main__":

    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
A bunch of handy functions for processing Hi-C data
(mainly in the form of matrices):

    -Normalizations
    -Interpolations
    -Filters
    -Removing artifacts
    -Quick sum-pooling (aka 'binning') in sparse and dense form
    -Simple models with parameter estimation
    -Computing best-matching 3D structures
    -Various metrics in use among Hi-C people for eyecandy purposes
     (directional index, domainograms, etc.)

These functions are meant to be simple and relatively quick
as-is implementations of procedures described in Hi-C papers.
"""

import numpy as np
import string
import collections
import itertools
import warnings


def despeckle_global(M, positions=None, stds=2):
    """Compute a trend by averaging all contacts of equal
    distance, then sets outstanding values (above stds standard
    deviations) to the expected value from the trend.
    """

    N = np.array(M, dtype=np.float64)
    n, m = M.shape
    if positions is None:
        positions = range(min(n, m))

    lengths = np.abs(np.diff(positions))

    def distance(i, j):
        mean_length = (lengths[i] + lengths[j]) / 2.
        if positions[i] < positions[j]:
            d = ((positions[j] - positions[i] - lengths[i]) + mean_length) / \
                1000.
        else:
            d = ((positions[i] - positions[j] - lengths[j]) + mean_length) / \
                1000.
        return d

    measurements, bins = {}, []
    for i in range(n):
        for j in range(1, i):
            d = distance(i, j)
            bins.append(np.abs(d))
            try:
                measurements[np.abs(d)].append(M[i, j])
            except KeyError:
                measurements[np.abs(d)] = [M[i, j]]

    mean = [np.mean(np.array(measurements[k])) for k in measurements.keys()]
    std = [np.std(np.array(measurements[k])) for k in measurements.keys()]

    for i, j in itertools.product(range(stds, n - stds),
                                  range(stds, m - stds)):
        d = distance(i, j)
        if M[i, j] >= mean[d] + std * stds:
            N[i, j] = mean[d]

    return (N + N.T) / 2


def despeckle_local(M, stds=2, width=2):
    """Replace outstanding values (above stds standard deviations)
    in a matrix by the average of a surrounding window of desired width.
    """

    N = np.array(M, dtype=np.float64)
    n, m = M.shape
    for i, j in itertools.product(range(width, n - width),
                                  range(width, m - width)):
        square = M[i - width:i + width, j - width:j + width]
        avg = np.average(square)
        std = np.std(square)
        if M[i, j] >= avg + stds * std:
            N[i, j] = avg
    return (N + N.T) / 2


def bin_dense(M, subsampling_factor=3):
    """Sum over each block of given subsampling factor, returns a matrix whose
    dimensions are this much as small (e.g. a 27x27 matrix binned with a
    subsampling factor equal to 3 will return a 9x9 matrix whose each component
    is the sum of the corresponding 3x3 block in the original matrix).
    Remaining columns and rows are summed likewise and added to the end of the
    new matrix.

    :note: Will not work for numpy verisons below 1.7
    """

    m = min(M.shape)
    n = (m // subsampling_factor) * subsampling_factor

    if n == 0:
        return np.array([M.sum()])

    N = np.array(M[:n, :n], dtype=np.float64)
    N = N.reshape(n // subsampling_factor, subsampling_factor,
                  n // subsampling_factor, subsampling_factor).sum(axis=(1, 3))
    if m > n:
        remaining_row = M[n:, :n]
        remaining_col = M[:n, n:]
        remaining_square = M[n:m, n:m]
        R = remaining_row.reshape(m % subsampling_factor,
                                  m // subsampling_factor,
                                  subsampling_factor).sum(axis=(0, 2))
        C = remaining_col.T.reshape(m % subsampling_factor,
                                    m // subsampling_factor,
                                    subsampling_factor).sum(axis=(0, 2)).T
        S = remaining_square.sum()
        N = np.append(N, [R], axis=0)
        result = np.append(N, np.array([list(C) + [S]]).T, axis=1)
    else:
        result = N

    return result


def bin_sparse(M, subsampling_factor=3):
    """Perform the bin_dense procedure for sparse matrices. Remaining rows
    and cols are lumped with the rest at the end.
    """

    try:
        from scipy.sparse import coo_matrix
    except ImportError as e:
        print(str(e))
        print("I am peforming dense binning by default.")
        return bin_dense(M.todense())

    N = M.tocoo()
    n, m = N.shape
    row, col, data = N.row, N.col, N.data

    # Divide row and column indices - duplicate coordinates are added in
    # sparse matrix construction

    binned_row = row // subsampling_factor
    binned_col = col // subsampling_factor
    binned_n = n // subsampling_factor
    binned_m = m // subsampling_factor

    # Attach remaining columns and rows to the last one

    binned_row[binned_row >= binned_n] -= n % subsampling_factor
    binned_col[binned_col >= binned_m] -= m % subsampling_factor

    result = coo_matrix((data, (binned_row, binned_col)),
                        shape=(binned_n, binned_m))
    return result


def bin_matrix(M, subsampling_factor=3):
    """Bin either sparse or dense matrices.
    """

    try:
        from scipy.sparse import issparse
        if issparse(M):
            return bin_sparse(M, subsampling_factor=subsampling_factor)
        else:
            raise ImportError
    except ImportError:
        return bin_dense(M, subsampling_factor=subsampling_factor)


def bin_annotation(annotation=None, subsampling_factor=3):
    """Perform binning on genome annotations such as contig information or bin
    positions.
    """

    if annotation is None:
        annotation = np.array([])
    n = len(annotation)
    binned_positions = [annotation[i] for i in range(n) if
                        i % subsampling_factor == 0]
    if len(binned_positions) == 0:
        binned_positions.append(0)
    return np.array(binned_positions)


def bin_measurement(measurement=None, subsampling_factor=3):
    """Perform binning on genome-wide measurements by summing each component
    in a window of variable size (subsampling_factor).
    """

    subs = int(subsampling_factor)
    if measurement is None:
        measurement = np.array([])
    n = len(measurement)
    binned_measurement = [measurement[i - subs + 1:i].sum()
                          for i in range(n) if i % subs == 0 and i > 0]
    return np.array(binned_measurement)


def build_pyramid(M, subsampling_factor=3):
    """Iterate over a given number of times on matrix M
    so as to compute smaller and smaller matrices with bin_dense.
    """

    subs = int(subsampling_factor)
    if subs < 1:
        raise ValueError(
            "Subsampling factor needs to be an integer greater than 1.")
    N = [M]
    while min(N[-1].shape) > 1:
        N.append(bin_matrix(N[-1], subsampling_factor=subs))
    return N


def bin_kb_dense(M, positions, length=10, contigs=None):
    """Perform binning with a fixed genomic length in
    kilobase pairs (kb). Fragments will be binned such
    that their total length is closest to the specified input.
    If a contig list is specified, binning will be performed
    such that fragments never overlap two contigs.
    """

    unit = 10**3
    ul = unit * length
    unit = positions / ul
    n = len(positions)
    idx = [i for i in range(
        n - 1) if np.ceil(unit[i]) < np.ceil(unit[i + 1])]
    binned_positions = positions[idx]
    m = len(idx) - 1
    N = np.zeros((m, m))
    for i in range(m):
        N[i] = np.array([M[idx[j]:idx[j + 1], idx[i]:idx[i + 1]].sum()
                        for j in range(m)])

    return N, binned_positions


def bin_exact_kb_dense(M, positions, length=10):
    """Perform the kb-binning procedure with total bin lengths being exactly
    set to that of the specified input. Fragments overlapping two potential
    bins will be split and related contact counts will be divided according
    to overlap proportions in each bin.
    """

    unit = 10**3
    ul = unit * length
    units = positions / ul
    n = len(positions)
    idx = [i for i in range(
        n - 1) if np.ceil(units[i]) < np.ceil(units[i + 1])]
    m = len(idx) - 1
    N = np.zeros((m, m))
    remainders = [0] + [np.abs(units[i] - units[i + 1]) for i in range(m)]
    for i in range(m):
        N[i] = np.array([(M[idx[j]:idx[j + 1], idx[i]:idx[i + 1]].sum() -
                         remainders[j] * M[i][j] +
                         remainders[j + 1] * M[i + 1][j])
                         for j in range(m)])
    return N


def bin_kb_sparse(M, positions, length=10):
    """Perform the exact kb-binning procedure on a sparse matrix.
    """

    try:
        from scipy.sparse import coo_matrix
    except ImportError as e:
        print(str(e))
        print("I am peforming dense normalization by default.")
        return bin_kb_dense(M.todense(), positions=positions)
    r = M.tocoo()
    unit = 10**3
    ul = unit * length
    units = positions / ul
    n = len(positions)
    indices = np.floor(units)
    row = [indices[np.floor(i)] for i in r.row / ul]
    col = [indices[np.floor(j)] for j in r.col / ul]
    binned_indices = positions[
        [i for i in range(n - 1) if np.ceil(units[i]) < np.ceil(units[i + 1])]]
    return coo_matrix((r.data, (row, col))), binned_indices


def trim_dense(M, n_std=3, s_min=None, s_max=None):
    """By default, return a matrix stripped of component
    vectors whose sparsity (i.e. total contact count on a
    single column or row) deviates more than specified number
    of standard deviations from the mean. Boolean variables
    s_min and s_max act as absolute fixed values which override
    such behaviour when specified.
    """

    M = np.array(M)
    sparsity = M.sum(axis=1)
    mean = np.mean(sparsity)
    std = np.std(sparsity)
    if s_min is None:
        s_min = mean - n_std * std
    if s_max is None:
        s_max = mean + n_std * std
    elif s_max == 0:
        s_max = np.amax(M)
    f = (sparsity > s_min) * (sparsity < s_max)
    N = M[f][:, f]
    return N


def trim_sparse(M, n_std=3, s_min=None, s_max=None):
    """Apply the trimming procedure to a sparse matrix.
    """

    try:
        from scipy.sparse import coo_matrix
    except ImportError as e:
        print(str(e))
        print("I am peforming dense normalization by default.")
        return trim_dense(M.todense())
    r = M.tocoo()
    sparsity = np.array(r.sum(axis=1)).flatten()
    mean = np.mean(sparsity)
    std = np.std(sparsity)
    if s_min is None:
        s_min = mean - n_std * std
    if s_max is None:
        s_max = mean + n_std * std
    f = (sparsity > s_min) * (sparsity < s_max)
    indices = [u for u in range(len(r.data)) if f[r.row[u]] and f[r.col[u]]]
    rows = np.array([r.row[i] for i in indices])
    cols = np.array([r.col[j] for j in indices])
    data = np.array([r.data[k] for k in indices])

    N = coo_matrix((data, (rows, cols)))
    return N


def normalize_dense(M, norm="frag", order=1, iterations=3):
    """Apply one of the many normalization types to input dense
    matrix. Will also apply any callable norms such as a user-made
    or a lambda function.
    """

    s = np.array(M, np.float64)
    floatorder = np.float64(order)

    if norm == "SCN":
        for _ in range(0, iterations):

            sumrows = s.sum(axis=1)
            maskrows = (sumrows != 0)[:, None] * (sumrows != 0)[None, :]
            sums_row = sumrows[:, None] * np.ones(sumrows.shape)[None, :]
            s[maskrows] = 1. * s[maskrows] / sums_row[maskrows]

            sumcols = s.sum(axis=0)
            maskcols = (sumcols != 0)[:, None] * (sumcols != 0)[None, :]
            sums_col = sumcols[None, :] * np.ones(sumcols.shape)[:, None]
            s[maskcols] = 1. * s[maskcols] / sums_col[maskcols]

    elif norm == "mirnylib":
        try:
            from mirnylib import numutils as ntls
            s = ntls.iterativeCorrection(s, iterations)[0]
        except ImportError as e:
            print(str(e))
            print("I can't find mirnylib.")
            print("Please install it from "
                  "https://bitbucket.org/mirnylab/mirnylib")
            print("I will use default norm as fallback.")
            return normalize_dense(M, order=order, iterations=iterations)

    elif norm == "frag":
        for _ in range(1, iterations):
            s_norm_x = np.linalg.norm(s, ord=floatorder, axis=0)
            s_norm_y = np.linalg.norm(s, ord=floatorder, axis=1)
            s_norm = np.tensordot(s_norm_x, s_norm_y, axes=0)
            s[s_norm != 0] = 1. * s[s_norm != 0] / s_norm[s_norm != 0]

    elif norm == "global":
        s_norm = np.linalg.norm(s, ord=floatorder)
        s /= 1. * s_norm

    elif callable(norm):
        s = norm(M)

    else:
        print("Unknown norm. Returning input as fallback")

    return (s + s.T) / 2


def normalize_sparse(M, norm="frag", order=1, iterations=3):
    """Applies a normalization type to a sparse matrix.
    """

    try:
        from scipy.sparse import csr_matrix
    except ImportError as e:
        print(str(e))
        print("I am peforming dense normalization by default.")
        return normalize_dense(M.todense())
    r = csr_matrix(M)
    if norm == "SCN":
        for _ in range(1, iterations):
            row_sums = np.array(r.sum(axis=1)).flatten()
            col_sums = np.array(r.sum(axis=0)).flatten()
            row_indices, col_indices = r.nonzero()
            r.data /= row_sums[row_indices] * col_sums[col_indices]

    elif norm == "global":
        try:
            from scipy.sparse import linalg
            r = linalg.norm(M, ord=order)
        except (ImportError, AttributeError) as e:
            print(str(e))
            print("I can't import linalg tools for sparse matrices.")
            print("Please upgrade your scipy version to 0.16.0.")

    elif callable(norm):
        r = norm(M)

    else:
        print("Unknown norm. Returning input as fallback")

    return r


def GC_partial(portion):
    """Manually compute GC content percentage in a DNA string, taking
    ambiguous values into account (according to standard IUPAC notation).
    """

    sequence_count = collections.Counter(portion)
    gc = ((sum([sequence_count[i] for i in 'gGcCsS']) +
          sum([sequence_count[i] for i in 'DdHh']) / 3.0 +
          2 * sum([sequence_count[i] for i in 'VvBb']) / 3.0 +
          sum([sequence_count[i] for i in 'NnYyRrKkMm']) / 2.0) / len(portion))
    return 0 or 100 * gc


def GC_wide(genome, window=1000):
    """Compute GC across a window of given length.

    :note: Requires Biopython
    """

    GC = []
    from Bio import SeqIO
    with open(genome) as handle:
        sequence = "".join([str(record.seq)
                            for record in SeqIO.parse(handle, "fasta")])

    n = len(sequence)
    for i in range(0, n, window):
        portion = sequence[i:min(i + window, n)]
        GC.append(GC_partial(portion))

    return GC


def split_genome(genome, chunk_size=10000):
    """Split genome into chunks of fixed size (save the last one).
    """

    chunks = []
    from Bio import SeqIO
    with open(genome) as handle:
        for record in SeqIO.parse(handle, "fasta"):
            sequence = record.seq
            n = len(sequence)
            chunks += [str(sequence[i:min(i + chunk_size, n)])
                       for i in range(0, n, chunk_size)]
    return np.array(chunks)


def directional(M, window=None, circ=False, extrapolate=True, log=True):
    """From a symmetrical matrix M of size n, return a vector d whose each
    component d[i] is a T-test of two samples represented by vectors of size
    window on either side of the i-th pixel on the diagonal. Edge elements may
    be extrapolated based on the vector size reduction, except in the case of
    circular genomes. If they aren't, d will be of size n - 2*(window-1)
    instead of n.
    """

    # Sanity checks
    if not type(M) is np.ndarray:
        M = np.array(M)

    if M.shape[0] != M.shape[1]:
        raise ValueError("Matrix is not square.")

    try:
        n = min(M.shape)
    except AttributeError:
        n = M.size

    # Default window argument
    if window is None:
        window = max(n // 100, 5)

    if window >= n:
        raise ValueError("Please choose a smaller window size.")

    try:
        from scipy.stats import ttest_rel
    except ImportError as e:
        print("Scipy not")
        print(str(e))
        raise
    if log:
        N = np.zeros((n, n))
        N[M > 0] = np.log(M[M > 0])
    else:
        N = M

    if circ:
        d = [ttest_rel(np.array(list(N[i, i - window:]) + list(N[i, :i])),
                       N[i, i:i + window])[0] for i in range(window)]
    elif extrapolate:
        d = [ttest_rel(N[i, 0:i], N[i, i:2 * i])[0] for i in range(window)]
    else:
        d = []

    d += [ttest_rel(N[i, i - window:i], N[i, i:i + window])[0]
          for i in range(window, n - window)]

    if circ:
        d += [ttest_rel(N[i, i - window:i],
                        np.array((list(N[i, :i - n + window]) +
                                  list(N[i, i:]))))[0]
              for i in range(n - window, n)]
    elif extrapolate:
        d += [ttest_rel(N[i, i - window:i],
                        (np.array(list(N[i, i:]) +
                         list(N[i, :window - (n - i)]))))[0]
              for i in range(n - window, n)]

    return d


def domainogram(M, window=None, circ=False, extrapolate=True):
    """From a symmetrical matrix M of size n, return a vector d whose each
    component d[i] is the total sum of a square of 2*window+1 size centered on
    the i-th main diagonal element. Edge elements may be extrapolated based on
    the square size reduction (i.e. for window = 4, the first component will be
    equal to the first diagonal pixel multiplied by 81, the second one will be
    equal to the first 2x2 square on the diagonal multiplied by 81/4, etc.),
    except in the case of circular genomes. If they aren't, d will be of size
    n - 2*(window-1) instead of n.
    """

    # Sanity checks
    if not type(M) is np.ndarray:
        M = np.array(M)

    if M.shape[0] != M.shape[1]:
        raise ValueError("Matrix is not square.")

    try:
        n = min(M.shape)
    except AttributeError:
        n = M.size

    # Default window argument
    if window is None:
        window = max(n // 100, 5)

    if window >= n:
        raise ValueError("Please choose a smaller window size.")

    if circ:
        d = [(np.sum(M[-i + window:, -i + window:]) +
             np.sum(M[:i - window + 1, :i - window + 1])
             for i in range(window))]
    elif extrapolate:
        d = [(np.sum(M[0:2 * i + 1, 0:2 * i + 1]) *
              ((2 * window + 1) ** 2.0) / ((2 * i + 1)**2.0))
             for i in range(window)]
    else:
        d = []

    d += [np.sum(M[i - window:i + window + 1, i - window:i + window + 1])
          for i in range(window, n - window)]

    if circ:
        d += [M[i:, i:].sum() + M[:n - i, n - i].sum()
              for i in range(n - window, n)]
    elif extrapolate:
        d += [M[i - window:, i - window:].sum() * ((2 * window + 1)**2.0) /
              ((2 * (n - i) + 1)**2.0) for i in range(n - window, n)]

    return np.array(d)


def from_dade_matrix(filename, header=False):
    """ Loads a numpy array from a Dade matrix instance,
    e.g.: A matrix containing the following (or equivalent in numpy)
        [['RST','chr1~0','chr1~10','chr2~0','chr2~30'],
        ['chr1~0','5', '10', '11', '2'],
        ['chr1~10',   '8', '3', '5'],
        ['chr2~0',         '3', '5'],
        ['chr2~30',             '5']]

    will return:
       [['5', '10', '11', '2'],
        ['10', '8', '3', '5'],
        [11', '3', '3', '5'],
        ['2', '5', '5', '5']], [chr1~0','chr1~10','chr2~0','chr2~30']

    Header data processing is delegated downstream.

    See https://github.com/scovit/DADE for more details about Dade.
    """

    A = np.genfromtxt(filename, delimiter="\t", dtype=None, filling_values=0)
    M, headers = np.array(A[1:, 1:], dtype=np.float64), A[0]
    matrix = M + M.T - np.diag(np.diag(M))
    parsed_header = list(
        zip(*[str(h)[:-1].strip('"').strip("'").split("~")
            for h in headers[1:]]))
    if header:
        return matrix, parsed_header
    else:
        return matrix


def to_dade_matrix(M, annotations="", filename=None):
    """Returns a Dade matrix from input numpy matrix. Any annotations are added
    as header. If filename is provided and valid, said matrix is also saved
    as text.
    """

    n, m = M.shape
    A = np.zeros((n + 1, m + 1))
    A[1:, 1:] = M
    if not annotations:
        annotations = np.array(["" for _ in n], dtype=str)
    A[0, :] = annotations
    A[:, 0] = annotations.T
    if filename:
        try:
            np.savetxt(filename, A, fmt='%i')
            print("I saved input matrix in dade format as " + str(filename))
        except ValueError as e:
            print("I couldn't save input matrix.")
            print(str(e))
        finally:
            return A

    return A


def from_structure(structure):
    """Return contact data from a 3D structure (in pdb format).
    """

    try:
        from Bio import PDB
        if isinstance(structure, str):
            p = PDB.PDBParser()
            structure = p.get_structure('S', structure)
        if isinstance(structure, PDB.Structure.Structure):
            for _ in structure.get_chains():
                atoms = [np.array(atom.get_coord())
                         for atom in structure.get_atoms()]
    except ImportError:
        print("Biopython not found.")
        raise

    atoms = np.array(structure)
    try:
        import scipy
        D = scipy.spatial.distance.pdist(atoms, 'euclidean')
        D = scipy.spatial.distance.squareform(D)
    except ImportError:
        print("Scipy not found.")
        raise
    m = np.max(1 / D[D != 0])
    M = np.zeros(D.shape)
    M[D != 0] = 1 / D[D != 0]
    M[D == 0] = m
    return M


def largest_connected_component(matrix):
    """Compute the adjacency matrix of the largest connected component of the
    graph whose input matrix is adjacent.
    """

    try:
        import scipy.sparse
        n, components = scipy.sparse.csgraph.connected_components(
            matrix, directed=False)
        print("I found " + str(n) + " connected components.")
        component_dist = collections.Counter(components)
        print("Distribution of components: " + str(component_dist))
        most_common, _ = component_dist.most_common(1)[0]
        ilcc = (components == most_common)
        return matrix[:, ilcc][ilcc]

    except ImportError as e:
        print("I couldn't find scipy which is needed for graph routines.")
        print(str(e))
        print("Returning input matrix as fallback.")
        return matrix


def to_structure(matrix, alpha=1):
    """Compute best matching 3D genome structure from underlying input matrix
    using ShRec3D-derived method from Lesne et al., 2014.

    Link: https://www.ncbi.nlm.nih.gov/pubmed/25240436

    The method performs two steps: first compute distance matrix by treating
    contact data as an adjacency graph (of weights equal to a power law
    function of the data), then embed the resulting distance matrix into
    3D space.

    The alpha parameter influences the weighting of contacts: if alpha < 1
    long-range interactions are prioritized; if alpha >> 1 short-range
    interactions have more weight wahen computing the distance matrix.
    """

    connected = largest_connected_component(matrix)
    distances = to_distance(connected, alpha)
    n, m = connected.shape
    bary = np.sum(np.triu(distances, 1)) / (n**2)  # barycenters
    d = np.array(np.sum(distances**2, 0) / n - bary)  # distances to origin
    gram = np.array([(d[i] + d[j] - distances[i][j]**2) / 2 for i,
                     j in itertools.product(range(n), range(m))]).reshape(n, m)
    normalized = gram / np.linalg.norm(gram, 'fro')

    try:
        symmetric = np.array((normalized + normalized.T) / 2,
                             dtype=np.longfloat)  # just in case
    except AttributeError:
        symmetric = np.array((normalized + normalized.T) / 2)

    from scipy import linalg
    eigen_values, eigen_vectors = linalg.eigh(symmetric)
    if not (eigen_values >= 0).all():
        warnings.warn("Negative eigen values were found.")
    idx = eigen_values.argsort()[-3:][::-1]
    values = eigen_values[idx]
    vectors = eigen_vectors[:, idx]
    coordinates = vectors * np.sqrt(values)
    return coordinates


def get_missing_bins(original, trimmed):
    """Retrieve indices of a trimmed matrix with respect to the original matrix.
    Fairly fast but is only correct if diagonal values are different, which is
    always the case in practice.
    """

    original_diag = np.diag(original)
    trimmed_diag = np.diag(trimmed)
    index = []
    m = min(original.shape)
    for j in range(min(trimmed.shape)):
        k = 0
        while original_diag[j + k] != trimmed_diag[j] and k < 2 * m:
            k += 1
        index.append(k + j)
    return np.array(index)


def to_pdb(structure, filename, contigs=None, annotations=None,
           indices=None, special_bins=None):
    """From a structure (or matrix) generate the corresponding pdb file
    representing each chain as a contig/chromosome and filling the occupancy
    field with a custom annotation. If the matrix has been trimmed somewhat,
    remaining indices may be specified.
    """

    n = len(structure)
    letters = (string.ascii_uppercase + string.ascii_lowercase +
               string.digits + string.punctuation) * int(n / 94 + 1)
    if contigs is None:
        contigs = np.ones(n + 1)
    if annotations is None:
        annotations = np.zeros(n + 1)
    if indices is None:
        indices = range(n + 1)
    if special_bins is None:
        special_bins = np.zeros(n + 1)

    structure_shapes_match = structure.shape[0] == structure.shape[1]
    if isinstance(structure, np.ndarray) and structure_shapes_match:
        structure = (to_structure(structure))

    X, Y, Z = (structure[:, i] for i in range(3))
    Xmax, Ymax, Zmax = (np.max(np.abs(Xi)) for Xi in (X, Y, Z))
    X *= 100.0 / Xmax
    Y *= 100.0 / Ymax
    Z *= 100.0 / Zmax

    X = np.around(X, 3)
    Y = np.around(Y, 3)
    Z = np.around(Z, 3)

    reference = ["OW", "OW", "CE", "TE", "tR"]

    with open(filename, 'w') as f:
        for i in range(1, n):
            line = "ATOM"  # 1-4 "ATOM"
            line += "  "  # 5-6 unused
            line += str(i).rjust(5)  # 7-11 atom serial number
            line += " "  # 12 unused
            line += reference[special_bins[i]].rjust(4)  # 13-16 atom name
            line += " "  # 17 alternate location indicator
            line += "SOL"  # 18-20 residue name
            line += " "  # 21 unused
            line += letters[int(contigs[indices[i]] - 1)
                            ]  # 22 chain identifier
            line += str(i).rjust(4)  # 23-26 residue sequence number
            line += " "  # 27 code for insertion of residues
            line += "   "  # 28-30 unused
            line += str(X[i]).rjust(8)  # 31-38 X orthogonal Å coordinate
            line += str(Y[i]).rjust(8)  # 39-46 Y orthogonal Å coordinate
            line += str(Z[i]).rjust(8)  # 47-54 Z orthogonal Å coordinate
            line += "1.00".rjust(6)  # 55-60 Occupancy
            # 61-66 Temperature factor
            line += str(annotations[i - 1]).rjust(6)
            line += "      "  # 67-72 unused
            line += "    "  # 73-76 segment identifier
            line += "O".rjust(2)  # 77-78 element symbol
            line += "\n"
            f.write(line)


def to_distance(matrix, alpha=1):
    """Compute distance matrix from contact data by applying a negative power
    law (alpha) to its nonzero pixels, then interpolating on the zeroes using a
    shortest-path algorithm.
    """
    matrix = np.array(matrix)
    try:
        import scipy.sparse
    except ImportError as e:
        print("Scipy not found.")
        print(str(e))
        raise

    if callable(alpha):
        distance_function = alpha
    else:
        try:
            a = np.float64(alpha)

            def distance_function(x):
                return 1 / (x ** (1 / a))
        except TypeError:
            print("Alpha parameter must be callable or an array-like")
            raise

    if hasattr(matrix, 'getformat'):
        distances = scipy.sparse.coo_matrix(matrix)
        distances.data = distance_function(distances.data)
    else:
        distances = np.zeros(matrix.shape)
        distances[matrix != 0] = distance_function(1 / matrix[matrix != 0])

    return scipy.sparse.csgraph.floyd_warshall(distances, directed=False)


def distance_to_contact(D, alpha=1):
    """Compute contact matrix from input distance matrix. Distance values of
    zeroes are given the largest contact count otherwise inferred non-zero
    distance values.
    """

    if callable(alpha):
        distance_function = alpha
    else:
        try:
            a = np.float64(alpha)

            def distance_function(x):
                return 1 / (x ** (1 / a))
        except TypeError:
            print("Alpha parameter must be callable or an array-like")
            raise
        except ZeroDivisionError:
            raise ValueError("Alpha parameter must be non-zero")

    m = np.max(distance_function(D[D != 0]))
    M = np.zeros(D.shape)
    M[D != 0] = distance_function(D[D != 0])
    M[D == 0] = m
    return M


def distance_law(matrix, log_bins=False):
    """Compute distance law as a function of the genomic coordinate aka P(s).
    Bin length increases exponentially with distance if log_bins is True.
    """

    D = np.array([np.average(np.diagonal(matrix, j))
                  for j in range(min(matrix.shape))])
    if not log_bins:
        return D
    else:
        n = min(matrix.shape)
        n_bins = int(np.log(n) / np.log(2) + 1)
        logD = np.array([np.average(D[int(2**(i - 1)):min(n, 2**i)])
                         for i in range(n_bins)])
        return logD


def shortest_path_interpolation(matrix, alpha=1, strict=True):
    """Perform interpolation on a matrix's data by using ShRec's shortest-path
    procedure backwards and forwards. This replaces zeroes with corresponding
    shortest-path based counts and may have the additional effect of 'blurring'
    the matrix somewhat. If strict is set to True, only zeroes are replaced
    this way.
    """
    matrix = np.array(matrix, np.float64)
    contacts = distance_to_contact(to_distance(matrix, alpha=alpha),
                                   alpha=alpha)
    if not strict:
        return contacts
    else:
        M = np.copy(matrix)
        M[matrix == 0] = contacts[matrix == 0]
        return M


def pdb_to_structure(filename):
    """Import a structure object from a PDB file.
    """

    try:
        from Bio.PDB import PDB
    except ImportError:
        print("I can't import Biopython which is needed to handle PDB files.")
        raise
    p = PDB.PDBParser()
    structure = p.get_structure('S', filename)
    for _ in structure.get_chains():
        atoms = [np.array(atom.get_coord()) for atom in structure.get_atoms()]
    return atoms


def noise(matrix):
    """Just a quick function to make a matrix noisy using a standard Poisson
    distribution (contacts are treated as rare events).
    """

    D = shortest_path_interpolation(matrix, strict=True)
    return np.random.poisson(lam=D)


def positions_to_contigs(positions):
    """Flattens and converts a positions array to a contigs array, if applicable.
    """

    if isinstance(positions, np.ndarray):
        flattened_positions = positions.flatten()
    else:
        try:
            flattened_positions = np.array(
                [pos for contig in positions for pos in contig])
        except TypeError:
            flattened_positions = np.array(positions)

    if (np.diff(positions) == 0).any() and not (0 in set(positions)):
        warnings.warn("I detected identical consecutive nonzero values.")
        return positions

    n = len(flattened_positions)
    contigs = np.ones(n)
    counter = 0
    for i in range(1, n):
        if positions[i] == 0:
            counter += 1
            contigs[i] += counter
        else:
            contigs[i] = contigs[i - 1]
    return contigs


def distance_diagonal_law(matrix, positions=None):
    """Compute a distance law trend using the contact averages of equal distances.
    Specific positions can be supplied if needed.
    """

    n = min(matrix.shape)
    if positions is None:
        return np.array([np.average(np.diagonal(matrix, j)) for j in range(n)])
    else:
        contigs = positions_to_contigs(positions)

    def is_intra(i, j):
        return contigs[i] == contigs[j]

    max_intra_distance = max((len(contigs == u) for u in set(contigs)))

    intra_contacts = []
    inter_contacts = [np.average(np.diagonal(matrix, j))
                      for j in range(max_intra_distance, n)]
    for j in range(max_intra_distance):
        D = np.diagonal(matrix, j)
        for i in range(len(D)):
            diagonal_intra = []
            if is_intra(i, j):
                diagonal_intra.append(D[i])
#            else:
#                diagonal_inter.append(D[i])
#        inter_contacts.append(np.average(np.array(diagonal_inter)))
        intra_contacts.append(np.average(np.array(diagonal_intra)))

    intra_contacts.extend(inter_contacts)

    return [positions, np.array(intra_contacts)]


def rippe_parameters(matrix, positions, lengths=None, init=None, circ=False):
    """Estimate parameters from the model described in Rippe et al., 2001.
    """

    n, _ = matrix.shape

    if lengths is None:
        lengths = np.abs(np.diff(positions))

    measurements, bins = [], []
    for i in range(n):
        for j in range(1, i):
            mean_length = (lengths[i] + lengths[j]) / 2.
            if positions[i] < positions[j]:
                d = (((positions[j] - positions[i] -
                     lengths[i]) + mean_length) /
                     1000.)
            else:
                d = (((positions[i] - positions[j] -
                     lengths[j]) + mean_length) /
                     1000.)

            bins.append(np.abs(d))
            measurements.append(matrix[i, j])
    parameters = estimate_param_rippe(measurements, bins, init=init, circ=circ)
    print(parameters)
    return parameters[0]


def estimate_param_rippe(measurements, bins, init=None, circ=False):
    """Perform least square optimization needed for the rippe_parameters function.
    """

    # Init values
    DEFAULT_INIT_RIPPE_PARAMETERS = [1., 9.6, -1.5]
    d = 3.

    def log_residuals(p, y, x):
        kuhn, lm, slope, A = p
        rippe = (np.log(A) + np.log(0.53) - 3 * np.log(kuhn) +
                 slope * (np.log(lm * x) - np.log(kuhn)) +
                 (d - 2) / ((np.power((lm * x / kuhn), 2) + d)))
        err = y - rippe

        return err

    def peval(x, param):

        if circ:
            l_cont = x.max()
            n = param[1] * x / param[0]
            n0 = param[1] * x[0] / param[0]
            n_l = param[1] * l_cont / param[0]
            s = n * (n_l - n) / n_l
            s0 = n0 * (n_l - n0) / n_l
            norm_lin = (param[3] * (0.53 * (param[0] ** -3.) *
                        np.power(n0, (param[2])) *
                        np.exp((d - 2) / ((np.power(n0, 2) + d)))))

            norm_circ = (param[3] * (0.53 * (param[0] ** -3.) *
                         np.power(s0, (param[2])) *
                         np.exp((d - 2) / ((np.power(s0, 2) + d)))))

            rippe = (param[3] * (0.53 * (param[0] ** -3.) *
                     np.power(s, (param[2])) * np.exp((d - 2) /
                     ((np.power(s, 2) + d)))) * norm_lin / norm_circ)

        else:

            rippe = (param[3] * (0.53 * (param[0] ** -3.) *
                     np.power((param[1] * x / param[0]), (param[2])) *
                     np.exp((d - 2) /
                     ((np.power((param[1] * x / param[0]), 2) + d)))))

        return rippe

    if init is None:
        init = DEFAULT_INIT_RIPPE_PARAMETERS

    A = np.sum(measurements)

    p0 = (p for p in init), A
    from scipy.optimize import leastsq
    plsq = leastsq(log_residuals, p0, args=(np.log(measurements), bins))

    y_estim = peval(bins, plsq[0])
    kuhn_x, lm_x, slope_x, A_x = plsq[0]
    plsq_out = [kuhn_x, lm_x, slope_x, d, A_x]

    np_plsq = np.array(plsq_out)

    if np.any(np.isnan(np_plsq)) or slope_x >= 0:
        warnings.warn("Problem in parameters estimation")
        plsq_out = p0

    return plsq_out, y_estim


def null_model(matrix, positions=None, lengths=None,
               model="uniform", noisy=False, circ=False,
               sparsity=False):
    """Attempt to compute a 'null model' of the matrix given a model
    to base itself on.
    """

    n, m = matrix.shape
    positions_supplied = True
    if positions is None:
        positions = range(n)
        positions_supplied = False
    if lengths is None:
        lengths = np.diff(positions)

    N = np.copy(matrix)

    contigs = np.array(positions_to_contigs(positions))

    def is_inter(i, j):
        return contigs[i] != contigs[j]

    diagonal = np.diag(matrix)

    if model == "uniform":
        if positions_supplied:
            trans_contacts = np.array([matrix[i, j]
                                      for i, j in itertools.product(range(n),
                                      range(m)) if is_inter(i, j)])
            mean_trans_contacts = np.average(trans_contacts)
        else:
            mean_trans_contacts = np.average(matrix) - diagonal / len(diagonal)

        N = np.random.poisson(lam=mean_trans_contacts, size=(n, m))
        np.fill_diagonal(N, diagonal)

    elif model == "distance":
        distances = distance_diagonal_law(matrix, positions)
        N = np.array([[distances[min(abs(i - j), n)]
                       for i in range(n)] for j in range(n)])

    elif model == "rippe":

        trans_contacts = np.array([matrix[i, j] for i, j in itertools.product(
            range(n), range(m)) if is_inter(i, j)])
        mean_trans_contacts = np.average(trans_contacts)
        kuhn, lm, slope, d, A = rippe_parameters(matrix, positions, circ=circ)

        def jc(s, frag):
            dist = s - circ * (s**2) / lengths[frag]
            computed_contacts = (0.53 * A * (kuhn**(-3.)) *
                                 (dist ** slope) *
                                 np.exp((d - 2) / (dist + d)))
            return np.maximum(computed_contacts, mean_trans_contacts)

        for i in range(n):
            for j in range(n):
                if not is_inter(i, j) and i != j:
                    posi, posj = positions[i], positions[j]
                    N[i, j] = jc(np.abs(posi - posj) * lm / kuhn, frag=j)
                else:
                    N[i, j] = mean_trans_contacts

    if sparsity:
        contact_sum = matrix.sum(axis=0)
        n = len(contact_sum)
        try:
            from Bio.Statistics import lowess
            trend = lowess(np.array(range(n), dtype=np.float64),
                           contact_sum, f=0.03)
        except ImportError:
            expected_size = int(np.amax(contact_sum) / np.average(contact_sum))
            w = min(max(expected_size, 20), 100)
            trend = np.array([np.average(contact_sum[i:min(i + w, n)]) for
                              i in range(n)])

        cov_score = np.sqrt((trend - np.average(trend)) / np.std(trend))

        N = ((N * cov_score).T) * cov_score

    if noisy:
        if callable(noisy):
            noise = noisy
        return noise(N)
    else:
        return N


def model_norm(matrix, positions=None,
               lengths=None, model="uniform", circ=False):

    N = null_model(matrix, positions, lengths, model,
                   noisy=False, circ=circ, sparsity=True)
    return matrix / shortest_path_interpolation(N, strict=True)


def trim_structure(struct, filtering="cube", n=2):
    """Remove outlier 'atoms' (aka bins) from a structure.
    """

    X, Y, Z = (struct[:, i] for i in range(3))

    if filtering == "sphere":
        R = (np.std(X)**2 + np.std(Y)**2 + np.std(Z)**2) * (n**2)
        f = (X - np.mean(X))**2 + (Y - np.mean(Y))**2 + (Z - np.mean(Z))**2 < R

    if filtering == "cube":
        R = min(np.std(X), np.std(Y), np.std(Z)) * n
        f = np.ones(len(X))
        for C in (X, Y, Z):
            f *= np.abs(C - np.mean(C)) < R

    if filtering == "percentile":
        f = np.ones(len(X))
        for C in (X, Y, Z):
            f *= np.abs(C - np.mean(C)
                        ) < np.percentile(np.abs(C - np.mean(C)), n)

    return np.array([X[f], Y[f], Z[f]])


def scalogram(M, circ=False):
    """Computes so-called 'scalograms' used to easily
    visualize contacts at different distance scales.
    Edge cases have been painstakingly taken
    care of.
    """

    # Sanity checks
    if not type(M) is np.ndarray:
        M = np.array(M)

    if M.shape[0] != M.shape[1]:
        raise ValueError("Matrix is not square.")

    try:
        n = min(M.shape)
    except AttributeError:
        n = M.size

    N = np.zeros(M.shape)
    for i in range(n):
        for j in range(n):
            if i + j < n and i >= j:
                N[i, j] = M[i, i - j:i + j + 1].sum()
            elif circ and i + j < n and i < j:
                N[i, j] = M[i, i - j:].sum() + M[i, :i + j + 1].sum()
            elif circ and i >= j and i + j >= n:
                N[i, j] = M[i, i - j:].sum() + M[i, :i + j - n + 1].sum()
            elif circ and i < j and i + j >= n:
                N[i, j] = (M[i, i - j:].sum() +
                           M[i, :].sum() +
                           M[i, :i + j - n + 1].sum())

    return N


def asd(M1, M2):
    """Compute a Fourier transform based distance
    between two matrices.

    Inspired from Galiez et al., 2015
    (https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4535829/)
    """

    from scipy.fftpack import fft2
    spectra1 = np.abs(fft2(M1))
    spectra2 = np.abs(fft2(M2))

    return np.linalg.norm(spectra2 - spectra1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
General utility functions for handling BAM files and generating 3C networks.
"""


import argparse
import collections
import copy
import itertools
import os
import numpy as np
import pysam
import gzip
import operator
from Bio import SeqIO
from Bio.SeqIO.QualityIO import FastqGeneralIterator
from Bio.Seq import Seq
from metator.scripts.log import logger

DEFAULT_SIZE_CHUNK_THRESHOLD = 500
DEFAULT_MAPQ_THRESHOLD = 10
DEFAULT_CHUNK_SIZE = 1000
DEFAULT_SELF_CONTACTS = False
DEFAULT_NORMALIZED = False
DEFAULT_READ_SIZE = 65
DEFAULT_NETWORK_FILE_NAME = "network.txt"
DEFAULT_CHUNK_DATA_FILE_NAME = "idx_contig_hit_size_cov.txt"

DEFAULT_PARAMETERS = {
    "size_chunk_threshold": DEFAULT_SIZE_CHUNK_THRESHOLD,
    "mapq_threshold": DEFAULT_MAPQ_THRESHOLD,
    "chunk_size": DEFAULT_CHUNK_SIZE,
    "read_size": DEFAULT_READ_SIZE,
    "self_contacts": DEFAULT_SELF_CONTACTS,
    "normalized": DEFAULT_NORMALIZED,
}


def alignment_to_contacts(
    sam_merged,
    assembly,
    output_dir,
    output_file_network=DEFAULT_NETWORK_FILE_NAME,
    output_file_chunk_data=DEFAULT_CHUNK_DATA_FILE_NAME,
    parameters=DEFAULT_PARAMETERS,
):
    """Generates a network file (in edgelist form) from an
    alignment in sam or bam format. Contigs are virtually split into
    'chunks' of nearly fixed size (by default between 500 and 1000 bp)
    to reduce size bias. The chunks are the network nodes and the edges
    are the contact counts.

    The network is in a strict barebone form so that it can be reused and
    imported quickly into other applications etc. Verbose information about
    every single node in the network is written on a 'chunk data' file,
    by default called 'idx_contig_hit_size_cov.txt'

    Parameters
    ----------
    sam_merged : file, str or pathlib.Path
        The alignment file in SAM/BAM format to be processed.
    assembly : file, str or pathlib.Path
        The initial assembly acting as the alignment file's reference genome.
    output_dir : str or pathlib.Path
        The output directory to write the network and chunk data into.
    output_dir_file_network : str or pathlib.Path, optional
        The specific file name for the output network file. Default is
        network.txt
    output_file_chunk_data : str or pathlib.Path, optional
        The specific file name for the output chunk data file. Default is
        idx_contig_hit_size_cov.txt
    parameters : dict, optional
        A dictionary of parameters for converting the alignment file into a
        network. These are:
        -size_chunk_threshold: the size (in bp) under which chunks are
        discarded. Default is 500.
        -mapq_threshold: the mapping quality under which alignments are
        discarded. Default is 10.
        -chunk_size: the default chunk size (in bp) when applicable, save
        smaller contigs or tail-ends. Default is 1000.
        -read_size: the size of reads used for mapping. Default is 65.
        -self_contacts: whether to count alignments between a chunk and
        itself. Default is False.
        -normalized: whether to normalize contacts by their coverage.
        Default is False.

    Returns
    -------
    chunk_complete_data : dict
        A dictionary where the keys are chunks in (contig, position) form and
        the values are their id, name, total contact count, size and coverage.
    all_contacts : dict
        A counter dictionary where the keys are chunk pairs and the values are
        their contact count.
    """

    all_contacts = collections.Counter()
    all_chunks = collections.Counter()

    #   Initialize parameters
    chunk_size = int(parameters["chunk_size"])
    mapq_threshold = int(parameters["mapq_threshold"])
    size_chunk_threshold = int(parameters["size_chunk_threshold"])
    read_size = int(parameters["read_size"])
    self_contacts = parameters["self_contacts"]
    normalized = parameters["normalized"]

    logger.info("Establishing chunk list...")
    chunk_complete_data = dict()

    #   Get all information about all chunks from all contigs
    #   (this gets updated at the end)
    global_id = 1
    for record in SeqIO.parse(assembly, "fasta"):
        length = len(record.seq)

        n_chunks = length // chunk_size
        n_chunks += (length % chunk_size) >= size_chunk_threshold

        for i in range(n_chunks):

            if (i + 1) * chunk_size <= length:
                size = chunk_size
            else:
                size = length % chunk_size

            chunk_name = "{}_{}".format(record.id, i)
            chunk_complete_data[chunk_name] = {
                "id": global_id,
                "hit": 0,
                "size": size,
                "coverage": 0,
            }
            global_id += 1

    logger.info("Opening alignment files...")

    current_read = None

    # Read the BAM file to detect contacts.
    with pysam.AlignmentFile(sam_merged, "rb") as alignment_merged_handle:

        names = alignment_merged_handle.references
        lengths = alignment_merged_handle.lengths
        names_and_lengths = {
            name: length for name, length in zip(names, lengths)
        }

        logger.info("Reading contacts...")

        # Since the BAM file is supposed to be sorted and interleaved,
        # pairs should be always grouped with one below the other (the exact
        # order doesn't matter since the network is symmetric, so we simply
        # treat the first one as 'forward' and the second one as 'reverse')

        # We keep iterating until two consecutive reads have the same name,
        # discarding ones that don't.

        while "Reading forward and reverse alignments alternatively":
            try:
                my_read = next(alignment_merged_handle)
                if current_read is None:
                    # First read
                    current_read = my_read
                    continue

                elif current_read.query_name != my_read.query_name:

                    # print("{}_{}".format(current_read, my_read))
                    current_read = my_read
                    continue

                read_forward, read_reverse = current_read, my_read

            except StopIteration:
                break

            # Get a bunch of info about the alignments to pass the tests below
            read_name_forward = read_forward.query_name
            read_name_reverse = read_reverse.query_name

            flag_forward, flag_reverse = read_forward.flag, read_reverse.flag

            try:
                assert read_name_forward == read_name_reverse
            except AssertionError:
                logger.error(
                    "Reads don't have the same name: " "%s and %s",
                    read_name_forward,
                    read_name_reverse,
                )
                raise

            # To check if a flag contains 4
            # (digit on the third position from the right in base 2),
            # 4 = unmapped in SAM spec
            def is_unmapped(flag):
                return np.base_repr(flag, padding=3)[-3] == "1"

            if is_unmapped(flag_forward) or is_unmapped(flag_reverse):
                # print("Detected unmapped read on one end, skipping")
                continue

            contig_name_forward = read_forward.reference_name
            contig_name_reverse = read_reverse.reference_name

            len_contig_for = names_and_lengths[contig_name_forward]
            len_contig_rev = names_and_lengths[contig_name_reverse]

            position_forward = read_forward.reference_start
            position_reverse = read_reverse.reference_start

            mapq_forward = read_forward.mapping_quality
            mapq_reverse = read_reverse.mapping_quality

            # Some more tests: checking for size, map quality, map status etc.
            mapq_test = min(mapq_forward, mapq_reverse) > mapq_threshold

            min_length = min(len_contig_for, len_contig_rev)
            length_test = min_length > size_chunk_threshold

            # Trickest test:
            #
            #
            #                contig
            #    pos1                          pos2
            #     ^                             ^
            # |-------|-------|-------|-------|---|
            # <-------><------><------><------><-->            <->
            #   chunk   chunk                  tail   size_chunk_threshold
            #
            # Test is passed if tail >= size_chunk_threshold (pos2)
            # or if the position is a non-tail chunk (pos1)

            if position_forward < chunk_size * (len_contig_for // chunk_size):
                current_chunk_forward_size = chunk_size
            else:
                current_chunk_forward_size = len_contig_for % chunk_size

            if position_reverse < chunk_size * (len_contig_rev // chunk_size):
                current_chunk_reverse_size = chunk_size
            else:
                current_chunk_reverse_size = len_contig_rev % chunk_size

            min_chunk_size = min(
                current_chunk_forward_size, current_chunk_reverse_size
            )

            chunk_test = min_chunk_size >= size_chunk_threshold

            if mapq_test and length_test and chunk_test:

                chunk_forward = position_forward // chunk_size
                chunk_reverse = position_reverse // chunk_size

                chunk_name_forward = "{}_{}".format(
                    contig_name_forward, chunk_forward
                )
                chunk_name_reverse = "{}_{}".format(
                    contig_name_reverse, chunk_reverse
                )

                if self_contacts or chunk_name_forward != chunk_name_reverse:

                    contact = tuple(
                        sorted((chunk_name_forward, chunk_name_reverse))
                    )

                    all_contacts[contact] += 1

                    chunk_key_forward = (
                        chunk_name_forward,
                        current_chunk_forward_size,
                    )
                    all_chunks[chunk_key_forward] += 1

                    chunk_key_reverse = (
                        chunk_name_reverse,
                        current_chunk_reverse_size,
                    )
                    all_chunks[chunk_key_reverse] += 1

    logger.info("Writing chunk data...")

    # Now we can update the chunk dictionary
    # with the info we gathered from the BAM file

    output_chunk_data_path = os.path.join(output_dir, output_file_chunk_data)

    with open(output_chunk_data_path, "w") as chunk_data_file_handle:

        for name in sorted(chunk_complete_data.keys()):

            chunk_data = chunk_complete_data[name]
            size = chunk_data["size"]
            chunk = (name, chunk_data["size"])
            hit = all_chunks[chunk]
            coverage = hit * read_size * 1.0 / size
            try:
                chunk_complete_data[name]["hit"] = hit
                chunk_complete_data[name]["coverage"] = coverage
            except KeyError:
                logger.error(
                    "A mismatch was detected between the reference "
                    "genome and the genome used for the alignment "
                    "file, some sequence names were not found"
                )
                raise

            idx = chunk_complete_data[name]["id"]
            line = "{}\t{}\t{}\t{}\t{}\n".format(
                idx, name, hit, size, coverage
            )
            chunk_data_file_handle.write(line)

    # Lastly, generate the network proper

    logger.info("Writing network...")

    output_network_path = os.path.join(output_dir, output_file_network)

    with open(output_network_path, "w") as network_file_handle:

        for chunks in sorted(all_contacts.keys()):

            chunk_name1, chunk_name2 = chunks
            contact_count = all_contacts[chunks]

            if normalized:
                coverage1 = chunk_complete_data[chunk_name1]["coverage"]
                coverage2 = chunk_complete_data[chunk_name2]["coverage"]
                mean_coverage = np.sqrt(coverage1 * coverage2)
                effective_count = contact_count * 1.0 / mean_coverage
            else:
                effective_count = contact_count

            try:
                idx1 = chunk_complete_data[chunk_name1]["id"]
                idx2 = chunk_complete_data[chunk_name2]["id"]
                line = "{}\t{}\t{}\n".format(idx1, idx2, effective_count)
                network_file_handle.write(line)
            except KeyError as e:
                logger.warning("Mismatch detected: %s", e)

    return chunk_complete_data, all_contacts


def merge_networks(output_file="merged_network.txt", *files):
    """Merge networks into a larger network.

    A naive implementation for merging two networks in edgelist format.

    Parameters
    ---------
    output_file : file, str, or pathlib.Path, optional
        The output file to write the merged network into. Default is
        merged_network.txt
    `*files` : file, str or pathlib.Path
        The network files to merge.

    Returns
    -------
    None

    Note
    ----
    The partitioning step doesn't mind redundant
    edges and handles them pretty well, so if you are not
    using the merged edgelist for anything else you can just
    concatenate the edgelists without using this function.
    """

    contacts = dict()
    for network_file in files:
        with open(network_file) as network_file_handle:
            for line in network_file_handle:
                id_a, id_b, n_contacts = line.split("\t")
                pair = sorted((id_a, id_b))
                try:
                    contacts[pair] += n_contacts
                except KeyError:
                    contacts[pair] = n_contacts

    sorted_contacts = sorted(contacts)
    with open(output_file, "w") as output_handle:
        for index_pair in sorted_contacts:
            id_a, id_b = index_pair
            n_contacts = contacts[index_pair]
            output_handle.write("{}\t{}\t{}\n".format(id_a, id_b, n_contacts))


def merge_chunk_data(output_file="merged_idx_contig_hit_size_cov.txt", *files):
    """Merge chunk data from different networks

    Similarly to merge_network, this merges any number of chunk data files.

    Parameters
    ---------
    output_file : file, str, or pathlib.Path, optional
        The output file to write the merged chunk data files into. Default is
        merged_idx_contig_hit_size_cov.txt
    `*files` : file, str or pathlib.Path
        The chunk data files to merge.

    Returns
    -------
    None
    """

    chunks = dict()
    for chunk_file in files:
        with open(chunk_file) as chunk_file_handle:
            for line in chunk_file_handle:
                chunk_id, chunk_name, hit, size, cov = line.split("\t")
                try:
                    chunks[chunk_id]["hit"] += hit
                    chunks[chunk_id]["cov"] += cov
                except KeyError:
                    chunks[chunk_id] = {
                        "name": chunk_name,
                        "hit": hit,
                        "size": size,
                        "cov": cov,
                    }

    sorted_chunks = sorted(chunks)
    with open(output_file, "w") as output_handle:
        for chunk_id in sorted_chunks:
            my_chunk = chunks[chunk_id]
            name, hit, size, cov = (
                my_chunk["name"],
                my_chunk["hit"],
                my_chunk["size"],
                my_chunk["cov"],
            )

            my_line = "{}\t{}\t{}\t{}\t{}".format(
                chunk_id, name, hit, size, cov
            )
            output_handle.write(my_line)


def alignment_to_reads(
    sam_merged,
    output_dir,
    parameters=DEFAULT_PARAMETERS,
    save_memory=True,
    *bin_fasta,
):
    """Generate reads from ambiguous alignment file

    Extract reads found to be mapping an input FASTA bin.
    If one read maps, the whole pair is extracted and written
    to the output paired-end FASTQ files. Reads that mapped
    and weren't part of a pair are kept in a third 'single'
    file for people who need it (e.g. to get extra paired reads
    by fetching the opposite one from the original FASTQ library).

    Parameters
    ----------
    sam_merged : file, str or pathlib.Path
        The input alignment file in SAM/BAM format to be processed.
    output_dir : str or pathlib.Path
        The output directory to write the network and chunk data into.
    parameters : dict, optional
        Parameters for the network to read conversion, similar to
        alignment_to_network.
    save_memory : bool, optional
        Whether to keep the read names into memory or write them in different
        files, which takes longer but may prevent out-of-memory crashes.
        Default is True.
    `*bin_fasta` : file, str or pathlib.Path
        The bin FASTA files with appropriately named records.

    Returns
    -------
    A dictionary of files with read names for each bin if save_memory is True,
    and a dictionary of the read names lists themselves otherwise.


    Note
    ----
    This will throw an IOError ('close failed in file object destructor') on
    exit with older versions of pysam for some reason. It's harmless but
    you may consider upgrading to a later version of pysam if it comes up in
    a pipeline.
    """

    #   Just in case file objects are sent as input
    def get_file_string(file_thing):
        try:
            file_string = file_thing.name
        except AttributeError:
            file_string = str(file_thing)
        return file_string

    #   Global set of chunks against which reads are required to
    #   map - we store them in a tuple that keeps track of the
    #   original bin each chunk came from so we can reattribute the reads later

    bin_chunks = set()
    for bin_file in bin_fasta:
        for record in SeqIO.parse(bin_file, "fasta"):
            bin_chunks.add((get_file_string(bin_file), record.id))

    chunk_size = int(parameters["chunk_size"])

    mapq_threshold = int(parameters["mapq_threshold"])

    def read_name(read):
        return read.query_name.split()[0]

    #   Since reading a huge BAM file can take up a
    #   lot of time and resources, we only do it once
    #   but that requires opening fastq files for writing
    #   as matching reads get detected along the
    #   bam and keeping track of which ones are
    #   currently open.

    def get_base_name(bin_file):

        base_name = ".".join(os.path.basename(bin_file).split(".")[:-1])

        output_path = os.path.join(
            output_dir, "{}.readnames".format(base_name)
        )

        return output_path

    if save_memory:
        opened_files = dict()
    else:
        read_names = dict()

    with pysam.AlignmentFile(sam_merged, "rb") as alignment_merged_handle:

        for (my_read_name, alignment_pool) in itertools.groupby(
            alignment_merged_handle, read_name
        ):

            for my_alignment in alignment_pool:

                relative_position = my_alignment.reference_start
                contig_name = my_alignment.reference_name

                chunk_position = relative_position // chunk_size

                # The 'chunk name' is used to detect macthing positions
                chunk_name = "{}_{}".format(contig_name, chunk_position)

                # But such matching positions have to map acceptably
                quality_test = my_alignment.mapping_quality > mapq_threshold

                for bin_file in bin_fasta:
                    chunk_tuple = (bin_file, chunk_name)
                    if chunk_tuple in bin_chunks and quality_test:
                        if save_memory:
                            output_path = get_base_name(bin_file)
                            try:
                                output_handle = opened_files[bin_file]
                            except KeyError:
                                output_handle = open(output_path, "w")
                                opened_files[bin_file] = output_handle

                            output_handle.write("@{}\n".format(my_read_name))

                        else:
                            try:
                                read_names[my_read_name].append(bin_file)
                            except KeyError:
                                read_names[my_read_name] = [bin_file]

    for file_handle in opened_files.values():
        file_handle.close()
    #   Return unpaired file names for pair_unpaired_reads() to process
    if save_memory:
        return opened_files.keys()
    else:
        return read_names


def retrieve_reads_from_fastq(
    fastq_forward, fastq_reverse, read_names, output_dir
):

    opened_files = dict()

    read_set = set(read_names.keys())

    def get_base_names(bin_file):

        base_name = ".".join(os.path.basename(bin_file).split(".")[:-1])

        for_fastq_path = os.path.join(
            output_dir, "{}_for.fastq".format(base_name)
        )
        rev_fastq_path = os.path.join(
            output_dir, "{}_rev.fastq".format(base_name)
        )

        return for_fastq_path, rev_fastq_path

    def seamless_open(my_file):

        try:
            return gzip.open(my_file)
        except IOError:
            return open(my_file)

    with seamless_open(fastq_forward) as forward_handle:
        with seamless_open(fastq_reverse) as reverse_handle:

            forward_fastq_iterator = FastqGeneralIterator(forward_handle)
            reverse_fastq_iterator = FastqGeneralIterator(reverse_handle)

            for read_for, read_rev in zip(
                forward_fastq_iterator, reverse_fastq_iterator
            ):

                name_for, sequence_for, quality_for = read_for
                name_rev, sequence_rev, quality_rev = read_rev
                short_name_for = name_for.split()[0]
                short_name_rev = name_rev.split()[0]

                if short_name_for in read_set or short_name_rev in read_set:
                    try:
                        bin_files = read_names[short_name_for]
                    except KeyError:
                        bin_files = read_names[short_name_rev]

                    for bin_file in bin_files:
                        (for_fastq_path, rev_fastq_path) = get_base_names(
                            bin_file
                        )

                        # Files are opened 'lazily' so as to not end up
                        # with a bunch of empty FASTQ files
                        try:
                            output_for_handle = opened_files[for_fastq_path]
                        except KeyError:
                            output_for_handle = open(for_fastq_path, "w")
                            opened_files[for_fastq_path] = output_for_handle

                        line_for = "@{}\n{}\n+\n{}\n".format(
                            name_for, sequence_for, quality_for
                        )
                        output_for_handle.write(line_for)

                        try:
                            output_rev_handle = opened_files[rev_fastq_path]
                        except KeyError:
                            output_rev_handle = open(rev_fastq_path, "w")
                            opened_files[rev_fastq_path] = output_rev_handle

                        line_rev = "@{}\n{}\n+\n{}\n".format(
                            name_rev, sequence_rev, quality_rev
                        )
                        output_rev_handle.write(line_rev)

    #   Close all the files we've opened for writing
    for file_handle in opened_files.values():
        file_handle.close()


def retrieve_reads_contig_wise(sam_merged, contig_data, output_dir):

    contig_dict = dict()
    opened_files = dict()

    def close_all_files():
        for my_file in opened_files.values():
            my_file.close()

    with open("contig_data.txt") as contig_data:
        for line in contig_data:
            fields = line.split()
            node = fields[0]
            core = fields[-3]
            contig_dict[node] = core

    query_getter = operator.attrgetter("query_name")

    with pysam.AlignmentFile(sam_merged, "rb") as sam_handle:
        for (my_read_name, alignment_pool) in itertools.groupby(
            sam_handle, query_getter
        ):
            my_read_set = dict()
            my_core_set = set()
            while "Reading alignments from alignment pool":
                try:
                    my_alignment = next(alignment_pool)
                    # print(contig_dict[my_alignment.reference_name])

                    is_reverse = my_alignment.is_reverse

                    my_seq = my_alignment.query_sequence

                    my_qual = my_alignment.query_qualities
                    my_qual_string = pysam.array_to_qualitystring(my_qual)

                    if is_reverse:
                        my_seq_string = str(Seq(my_seq).reverse_complement())
                        my_qual_string = my_qual_string[::-1]
                    else:
                        my_seq_string = my_seq

                    my_seq_tuple = (my_seq_string, my_qual_string)

                    if len(my_read_set) > 2:
                        logger.warning(
                            "Something's gone wrong with read set %s, as "
                            "there are %s of them",
                            my_read_name,
                            len(my_read_set),
                        )
                    elif len(my_read_set) == 0:
                        my_read_set[my_seq_tuple] = "forward"
                    elif my_seq_tuple not in my_read_set.keys():
                        my_read_set[my_seq_tuple] = "reverse"
                    try:
                        ref = contig_dict[my_alignment.reference_name]
                        my_core_set.add(ref)
                    except KeyError:
                        my_core_set.add("unknown")
                except StopIteration:
                    if len(my_read_set) == 2:
                        for core_name in my_core_set:
                            for my_tuple, file_id in my_read_set.items():
                                if file_id == "forward":
                                    file_end = ".end1"
                                elif file_id == "reverse":
                                    file_end = ".end2"

                                basename = "{core_name}{file_end}".format(
                                    core_name, file_end
                                )
                                filename = os.path.join(output_dir, basename)
                                try:
                                    file_to_write = opened_files[filename]
                                except KeyError:
                                    file_to_write = open(filename, "w")
                                    opened_files[filename] = file_to_write
                                except IOError:
                                    logger.error(
                                        "Error when trying to handle"
                                        "%s. Maybe there are too many opened"
                                        "files at once: %s",
                                        filename,
                                        len(opened_files),
                                    )
                                    raise

                                seq, qual = my_tuple
                                line = "@{}\n{}\n+\n{}\n".format(
                                    my_read_name, seq, qual
                                )
                                file_to_write.write(line)
                    elif len(my_read_set) == 1:
                        for core_name in my_core_set:
                            file_end = ".end"
                            basename = "{}{}".format(core_name, file_end)
                            filename = os.path.join(output_dir, basename)
                            try:
                                file_to_write = opened_files[filename]
                            except KeyError:
                                file_to_write = open(filename, "w")
                                opened_files[filename] = file_to_write
                            except IOError:
                                logger.error(
                                    "Error when trying to handle"
                                    "%s. Maybe there are too many opened"
                                    "files at once: %s",
                                    filename,
                                    len(opened_files),
                                )
                                raise
                            seq, qual = my_read_set.keys()[0]
                            line = "@{}\n{}\n+\n{}\n".format(
                                my_read_name, seq, qual
                            )
                            file_to_write.write(line)
                    else:
                        logger.warning(
                            "Something's gone wrong with read set %s, as "
                            "there are %s of them",
                            my_read_name,
                            len(my_read_set),
                        )
                    break

    close_all_files()


def main():

    parser = argparse.ArgumentParser(description="Process alignment files.")

    parser.add_argument(
        "-i",
        "--input",
        type=argparse.FileType("r"),
        help="Merged (interlaced) alignment file",
        required=True,
    )

    parser.add_argument(
        "-f", "--reference", help="Reference fasta file", nargs="+"
    )

    parser.add_argument(
        "-o", "--output", help="Output directory", required=True
    )

    parser.add_argument(
        "-q",
        "--map-quality",
        type=int,
        help="Minimum mapping quality threshold",
        default=DEFAULT_PARAMETERS["mapq_threshold"],
    )

    parser.add_argument(
        "-c",
        "--chunk-size",
        type=int,
        help="Standard chunk size for nodes in the network",
        default=DEFAULT_PARAMETERS["chunk_size"],
    )

    parser.add_argument(
        "-s",
        "--size-chunk-threshold",
        type=int,
        help="Minimum size for tail ends to be integrated"
        " as chunks in the network",
        default=DEFAULT_PARAMETERS["size_chunk_threshold"],
    )

    parser.add_argument(
        "-a",
        "--self-contacts",
        action="store_true",
        help="Do not discard self contacts",
        default=DEFAULT_PARAMETERS["self_contacts"],
    )

    parser.add_argument(
        "-n",
        "--normalize",
        action="store_true",
        help="Normalize contacts by the geometric mean"
        " of both coverages of the chunks",
        default=DEFAULT_PARAMETERS["normalized"],
    )

    parser.add_argument(
        "-r",
        "--read-size",
        help="Read size",
        default=DEFAULT_PARAMETERS["read_size"],
    )

    parser.add_argument(
        "-F",
        "--fastq",
        help="Reconstruct FASTQ from "
        "reference and interleaved alignment file",
        nargs=2,
    )

    parser.add_argument(
        "-D",
        "--fastq-contig",
        help="Reconstruct FASTQ from "
        "contig data and interleaved alignment file",
        nargs=1,
    )

    parser.add_argument(
        "-m",
        "--mem",
        help="Save memory by only writing " " read names in FASTQ mode",
        action="store_true",
    )

    args = parser.parse_args()

    merged_file = args.input
    reference_file = args.reference
    output_dir = args.output
    fastq_files = args.fastq
    contig_data = args.fastq_contig
    save_memory = args.mem

    parameters = copy.deepcopy(DEFAULT_PARAMETERS)

    parameters["mapq_threshold"] = args.map_quality
    parameters["chunk_size"] = args.chunk_size
    parameters["read_size"] = args.read_size
    parameters["size_chunk_threshold"] = args.size_chunk_threshold
    parameters["self_contacts"] = args.self_contacts
    parameters["normalized"] = args.normalize

    if fastq_files:

        fastq_forward, fastq_reverse = fastq_files
        read_names = alignment_to_reads(
            merged_file, output_dir, parameters, save_memory, *reference_file
        )

        if save_memory:
            retrieve_reads_from_fastq(
                fastq_forward, fastq_reverse, read_names, output_dir
            )

    elif contig_data:
        retrieve_reads_contig_wise(merged_file, contig_data, output_dir)

    else:

        my_assembly, _ = reference_file

        alignment_to_contacts(
            sam_merged=merged_file,
            assembly=my_assembly,
            output_dir=output_dir,
            parameters=parameters,
        )


if __name__ == "__main__":

    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Quickly draw figures from metaTOR calls.
"""

import matplotlib

matplotlib.use("Agg")

import argparse

import numpy as np

import matplotlib.backends.backend_pdf
from metator.scripts import hicstuff as hcs
from matplotlib import pyplot as plt
from scipy import sparse


# Use (prettier) seaborn if available
SEABORN = False

try:
    import seaborn as sns

    SEABORN = True
except ImportError:
    pass

DEFAULT_INTERVALS = (0, 2, 10, 50, 100, 250, 500, 1000, np.inf)
DEFAULT_SATURATION_THRESHOLD = 80
DEFAULT_MAX_SIZE_MATRIX = 2000
DEFAULT_DPI = 200


def spaceless_pdf_plot_maker(array, filename, vmax=None, dpi=DEFAULT_DPI):
    """Draw a pretty plot from an array

    A function that performs all the tedious matplotlib
    magic to draw a 2D array with as few parameters and
    as little whitespace as possible.

    Parameters
    ----------
    array : array_like
        The input array to draw.
    filename : file, str or pathlib.Path
        The output image to save the array into.
    vmax : float, optional
        The default saturation threshold for the array. If set to None, the
        80th percentile value of the array is chosen. Default is None.
    dpi : int, optional
        Dots per inch (DPI) of the output image. Default is 200.

    Returns
    -------
    None
    """

    if vmax is None:
        vmax = np.percentile(array, DEFAULT_SATURATION_THRESHOLD)
    plt.gca().set_axis_off()
    plt.subplots_adjust(top=1, bottom=0, right=1, left=0, hspace=0, wspace=0)
    plt.margins(0, 0)
    plt.gca().xaxis.set_major_locator(plt.NullLocator())
    plt.gca().yaxis.set_major_locator(plt.NullLocator())
    plt.figure()
    if SEABORN:
        sns.heatmap(array, vmax=vmax, cmap="Reds")
    else:
        plt.imshow(array, vmax=vmax, cmap="Reds", interpolation="none")
        plt.colorbar()
    plt.savefig(filename, bbox_inches="tight", pad_inches=0.0, dpi=dpi)
    plt.close()


def draw_sparse_matrix(
    array_filename,
    output_image,
    vmax=DEFAULT_SATURATION_THRESHOLD,
    max_size_matrix=DEFAULT_MAX_SIZE_MATRIX,
):
    """Draw a quick preview of a sparse matrix with automated
    binning and normalization.
    """

    matrix = np.loadtxt(array_filename, dtype=np.int32, skiprows=1)
    try:
        row, col, data = matrix.T
    except ValueError:
        row, col, data = matrix
    size = max(np.amax(row), np.amax(col)) + 1
    S = sparse.coo_matrix((data, (row, col)), shape=(size, size))
    if max_size_matrix <= 0:
        binning = 1
    else:
        binning = (size // max_size_matrix) + 1
    binned_S = hcs.bin_sparse(S, subsampling_factor=binning)
    dense_S = binned_S.todense()
    dense_S = dense_S + dense_S.T - np.diag(np.diag(dense_S))
    normed_S = hcs.normalize_dense(dense_S)
    spaceless_pdf_plot_maker(normed_S, output_image, vmax=vmax)


def make_barplots(sizes_file, output, intervals=DEFAULT_INTERVALS):

    sizes = np.loadtxt(sizes_file)

    data_for_barplot = [sum(sizes)]

    n = len(intervals) - 1
    for i in range(n):
        fits_interval = (sizes > intervals[i]) * (sizes < intervals[i + 1])
        chunks_in_interval = sizes[fits_interval].sum()
        data_for_barplot.append(chunks_in_interval)

    labels = ["Total"]

    for i in range(n):
        if intervals[i + 1] == np.inf:
            labels.append("> {}".format(intervals[i]))
        elif intervals[i] == 0:
            labels.append("1")
        else:
            labels.append("{} - {}".format(intervals[i], intervals[i + 1]))

    if SEABORN:
        sns.barplot(labels, data_for_barplot)
    else:
        y = np.arange(n + 1)
        plt.bar(y, data_for_barplot, align="center", alpha=0.5)
        plt.xticks(y, labels)

    plt.xlabel("Size of parent core (in chunks)")
    plt.ylabel("Number of chunks")
    plt.title("Distribution of chunks by size of parent core")
    plt.savefig(output, bbox_inches="tight", pad_inches=0.0)


def draw_regression(sizes_file, output):

    try:
        size_label = "{}".format(sizes_file.split(".")[0].split("_")[-1])
    except ValueError:
        size_label = ""

    sizes_array = np.loadtxt(sizes_file, usecols=(0, 1))
    indices = sizes_array[:, 0]
    values = sizes_array[:, 1]

    if SEABORN:
        sns.regplot(indices, values, fit_reg=False)
    else:
        plt.scatter(indices, values)

    plt.xlabel("Number of iterations")
    plt.ylabel("Number of bins")

    plt.title("Evolution of bins > {}".format(size_label))
    plt.savefig(output, bbox_inches="tight", pad_inches=0.0)


def draw_enrichments(output, *files):

    sizes = []
    labels = []

    n = len(files)

    all_files = list(files)

    for enrichment_file in all_files:
        data_enrichments = np.loadtxt(enrichment_file, usecols=(1, 2))
        label = enrichment_file.split("/")[-1].split("_")[0]

        my_size = data_enrichments[:, 0]
        my_hits = data_enrichments[:, 1]

        my_total_hits = np.int32(np.sum(my_hits))

        unweighted_sizes = np.zeros(my_total_hits)

        cursor = 0
        for u, v in zip(my_size, my_hits):
            slice_size = np.int32(v)
            unweighted_sizes[cursor : cursor + slice_size] = u
            cursor += slice_size

        sizes.append(np.log10(unweighted_sizes[unweighted_sizes > 1]))
        labels.append(label)

    if SEABORN:
        sizes = [size.tolist() for size in sizes]
        sns.violinplot(
            data=sizes,
            width=1,
            inner="box",
            cut=0,
            linewidth=.3,
            bw=.18,
            color="grey",
        )
    else:
        labels = [""] + labels
        plt.violinplot(
            sizes,
            showmeans=False,
            showmedians=False,
            showextrema=False,
            widths=1,
            bw_method=.18,
        )

    y = np.arange(n)

    plt.xticks(y, labels)

    plt.xlabel("Number of hits")
    plt.ylabel("log(Sizes)")

    plt.title("Distribution of hits vs bin sizes")
    plt.savefig(output, bbox_inches="tight", pad_inches=0.0)


def draw_logplots(enrichment_files, output):

    proportional = "proportion" in enrichment_files or "proportion" in output

    enrichment_data = np.loadtxt(enrichment_files, usecols=(1, 2))
    sizes = enrichment_data[:, 0]
    hits = enrichment_data[:, 1]

    if SEABORN:
        sns.regplot(sizes, hits, fit_reg=False)
    else:
        plt.scatter(sizes, hits)

    plt.xlabel("Bin size")
    plt.ylabel("Number of hits")

    if proportional:
        plt.xscale("log")
        plt.title("Bin size vs relative hits")
    else:
        plt.loglog()
        plt.title("Bin size vs hits")
    plt.savefig(
        "{}_nhit.pdf".format(output), bbox_inches="tight", pad_inches=0.0
    )
    plt.close()


def main():

    parser = argparse.ArgumentParser(
        description="Draw and save matrices," "barplots, boxplots and so on."
    )

    parser.add_argument(
        "-p", "--plots", help="Draw scatterplots for core size evolution"
    )

    parser.add_argument("-m", "--matrices", help="Draw contact maps")

    parser.add_argument(
        "-b",
        "--barplots",
        help="Draw barplots for" "parent core size distribution",
    )

    parser.add_argument(
        "-v",
        "--violins",
        help="Draw violin plots to view" "hits vs. size distribution",
        nargs="*",
    )

    parser.add_argument(
        "-l",
        "--logplots",
        help="Draw log plots to view"
        "bin size distribution for different hits",
    )

    parser.add_argument("-s", "--sparse", help="Draw sparse matrix")

    parser.add_argument(
        "-o",
        "--output",
        required=True,
        help="Output image file to draw in any extension",
    )

    args = parser.parse_args()

    plots = args.plots
    matrices = args.matrices
    barplots = args.barplots
    logplots = args.logplots
    violins = args.violins
    output_file = args.output
    sparse_matrix = args.sparse

    if barplots:
        input_file = barplots
        make_barplots(input_file, output_file)

    if matrices:
        input_file = matrices
        spaceless_pdf_plot_maker(input_file, output_file)

    if plots:
        input_file = plots
        draw_regression(input_file, output_file)

    if logplots:
        input_file = logplots
        draw_logplots(input_file, output_file)

    if violins:
        input_files = violins
        draw_enrichments(output_file, *input_files)

    if sparse_matrix:
        input_file = sparse_matrix
        draw_sparse_matrix(input_file, output_file)


if __name__ == "__main__":

    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
A few handy functions for processing fasta files.
"""

from Bio import SeqIO
import argparse
import re

try:
    from metator.scripts.network import DEFAULT_CHUNK_SIZE
except ImportError:
    DEFAULT_CHUNK_SIZE = 1000


def rename_genome(genome_in, genome_out=None):
    """Rename genome and slugify headers

    Rename genomes according to a simple naming scheme; this
    is mainly done to avoid special character weirdness.

    Parameters
    ----------
    genome_in : file, str or pathlib.Path
        The input genome to be renamed and slugify.
    genome_out : file, str or pathlib.Path
        The output genome to be written into. Defaults is <base>_renamed.fa,
        where <base> is genome_in without its extension.

    Returns
    -------
    None
    """

    if genome_out is None:
        genome_out = "{}_renamed.fa".format(genome_in.split(".")[0])

    with open(genome_out, "w") as output_handle:
        for record in SeqIO.parse(genome_in, "fasta"):

            # Replace hyphens, tabs and whitespace with underscores
            new_record_id = record.id.replace(" ", "_")
            new_record_id = new_record_id.replace("-", "_")
            new_record_id = new_record_id.replace("\t", "_")

            # Remove anything that's weird, i.e. not alphanumeric
            # or an underscore
            new_record_id = re.sub("[^_A-Za-z0-9]+", "", new_record_id)
            header = ">{}\n".format(new_record_id)

            output_handle.write(header)
            output_handle.write("{}\n".format(str(record.seq)))


def filter_genome(genome_in, threshold=500, list_records=None):
    """Filter fasta file according to various parameters.

    Filter a fasta file according to size and/or an explicit list of records
    to keep.

    Parameters
    ----------
    genome_in: file, str or pathlib.Path
        The input genome in FASTA format.
    threshold: int, optional
        The size below which genome records are discarded. Default is the
        default minimum chunk size, i.e. 500.
    list_records: array_like, optional
        A list of record ids to keep. If not None, records that don't belong
        to that list are discarded. Default is None, i.e. all records are
        kept.

    Returns
    -------
    records_to_write: generator
        Filtered records that were kept.
    """

    if list_records is None:

        def truth(*args):
            del args
            return True

        is_a_record_to_keep = truth

    else:
        try:
            with open(list_records) as records_handle:
                records_to_keep = records_handle.readlines()
        except OSError:
            if not hasattr(list_records, "__contains__"):
                raise
            else:
                records_to_keep = list_records

        is_a_record_to_keep = records_to_keep.__contains__

    records_to_write = (
        record
        for record in SeqIO.parse(genome_in, "fasta")
        if (len(record.seq) >= threshold and is_a_record_to_keep(record.id))
    )

    return records_to_write


def rename_proteins(prot_in, prot_out=None, chunk_size=DEFAULT_CHUNK_SIZE):
    """Rename prodigal output files

    Rename output files from prodigal according to the following naming
    scheme: >contigX_chunkY__geneZ

    Chunk numbering starts at 0 and gene identification is taken from prodigal.

    Parameters
    ----------
    prot_in : file, str or pathlib.Path
        The input protein file in FASTA format to be renamed.
    prot_out : file, str or pathlib.Path
        The output protein file to be renamed into.
    chunk_size : int, optional
        The size of the chunks (in bp) used in the pipeline. Default is 1000.

    Returns
    -------
    None
    """

    if prot_out is None:
        prot_out = "{}_renamed.fa".format(prot_in.split(".")[0])

    with open(prot_out, "w") as prot_out_handle:

        for record in SeqIO.parse(prot_in, "fasta"):
            header = record.description
            name, pos_start, _, _, _ = header.split("#")

            chunk_start = int(pos_start) // chunk_size

            name_split = name.split("_")
            contig_name = "_".join(name_split[:-1])
            gene_id = name_split[-1]

            new_record_id = "{}_{}__gene{}".format(
                contig_name, chunk_start, gene_id
            )

            prot_out_handle.write(">{}\n".format(new_record_id))
            prot_out_handle.write("{}\n".format(str(record.seq)))


def write_records(records, output_file, split=False):

    """Write FASTA records

    Write a FASTA file from an iterable of records.

    Parameters
    ----------
    records : iterable
        Input records to write.
    output_file : file, str or pathlib.Path
        Output FASTA file to be written into.
    split : bool, optional
        If True, each record is written into its own separate file. Default is
        False.

    Returns
    -------
    None
    """

    if split:
        for record in records:
            with open(
                "{}{}.fa".format(output_file, record.id), "w"
            ) as record_handle:
                SeqIO.write(record, record_handle, "fasta")
    else:
        SeqIO.write(records, output_file, "fasta")


def main():

    parser = argparse.ArgumentParser(
        description="A few utilities"
        " to manipulate "
        "and filter fasta files."
    )
    parser.add_argument(
        "-i", "--input", help="Input fasta file", required=True
    )
    parser.add_argument("-o", "--output", help="Output file", required=True)

    parser.add_argument(
        "-t",
        "--threshold",
        type=int,
        help="Threshold size of records to keep",
        default=0,
    )

    parser.add_argument(
        "-r", "--records", help="List of records to keep", default=None
    )

    parser.add_argument(
        "-s",
        "--split",
        help="Split input file into single record fasta files",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "-n",
        "--rename",
        help="Rename fasta file ids " "and remove special characters",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-p",
        "--proteins",
        help="Rename prodigal output file",
        action="store_true",
        default=False,
    )

    args = parser.parse_args()

    fasta_file = args.input
    output_file = args.output
    threshold = args.threshold
    records = args.records
    split = args.split
    rename = args.rename
    proteins = args.proteins

    if proteins:
        rename_proteins(prot_in=fasta_file, prot_out=output_file)

    elif rename:
        rename_genome(genome_in=fasta_file, genome_out=output_file)

    else:
        records_to_write = filter_genome(
            genome_in=fasta_file, threshold=threshold, list_records=records
        )

        write_records(
            records=records_to_write, output_file=output_file, split=split
        )


if __name__ == "__main__":

    main()



#!/usr/bin/env python3

"""Basic logging setup for metaTOR.

Logging level can be set by the user and determines the verbosity of the
whole program.
"""


import logging

from logging import FileHandler

# Log level is modified by other modules according to the entry (user set)
# value
CURRENT_LOG_LEVEL = logging.INFO

logging.captureWarnings(True)

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)

logfile_formatter = logging.Formatter(
    "%(asctime)s :: %(levelname)s :: %(message)s"
)
stdout_formatter = logging.Formatter("%(levelname)s :: %(message)s")

file_handler = FileHandler("metator.log", "a")

file_handler.setLevel(logging.INFO)
file_handler.setFormatter(logfile_formatter)
logger.addHandler(file_handler)

errorfile_handler = FileHandler("metator.err", "a")
errorfile_handler.setLevel(logging.ERROR)
errorfile_handler.setFormatter(logfile_formatter)
logger.addHandler(errorfile_handler)

stream_handler = logging.StreamHandler()
stream_handler.setLevel(CURRENT_LOG_LEVEL)
stream_handler.setFormatter(stdout_formatter)
logger.addHandler(stream_handler)
